// Copyright (C) 2017 Ben Asselstine
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <argp.h>
#include <string.h>
#include <stdlib.h>
#include "xges-create-priv.h"
#include "clip-priv.h"
#include <unistd.h>

//instead of pulling in gstreamer
static const unsigned long long int gst_second = 1000000000;

void
init_clip (clip_t *clip, int layer)
{
  size_t length = sizeof (struct clip_details_t);
  *clip = calloc (1, length);
  struct clip_details_t *c = (struct clip_details_t*) *clip;
  c->layer_number = layer;
  c->position = -1;
  c->volume = -1;
  c->greenscreen_threshold = 100;
  c->greenscreen_distance = 28.799999999999998;
  return;
}

void
init_clip_from_clip (clip_t *clip, clip_t clip_to_copy_from)
{
  init_clip (clip, -1);
  struct clip_details_t *c = (struct clip_details_t*) clip_to_copy_from;
  memcpy (*clip, c, sizeof (struct clip_details_t));
  struct clip_details_t *nc = (struct clip_details_t*) *clip;

  if (c->cached_uri)
    nc->cached_uri = strdup (c->cached_uri);

}

void
destroy_clip (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  if (c->cached_uri)
    free (c->cached_uri);
  free (clip);
  return;
}

void
insert_clip_after (clip_t clip, int idx, clip_t **clips, int *num_clips)
{
  int i;
  (*num_clips)++;
  *clips = realloc (*clips, sizeof (clip_t) * *num_clips);
  for (i = *num_clips - 2; i > idx; i--)
    (*clips)[i+1] = (*clips)[i];
  (*clips)[idx+1] = clip;
}

static error_t
parse_clip_opt (int key, char *arg, struct argp_state *state)
{
  struct xges_create_details_t *xges_create = (struct xges_create_details_t*) state->input;
  char *end;
  if (xges_create->num_clips == 0 && !is_a_clip_option (key))
    return 0;
  else if (xges_create->num_clips == 0 && is_a_clip_option (key))
    argp_error (state, "you must specify `" CLIP_CMD " ' before using one of its options");
  else
    check_bad_option (key, state);
  struct clip_details_t *c = (struct clip_details_t*) xges_create->clips[xges_create->num_clips-1];
  switch (key)
    {
    case CLIP_START:
        {
          double val = strtod (arg, &end);
          if (*end != '\0' || val < 0)
            argp_error (state, "`%s' is a bad value for --start", arg);
          else
            c->start = val;
        }
      break;
    case CLIP_DURATION:
        {
          double val = strtod (arg, &end);
          if (*end != '\0' || val <= 0)
            argp_error (state, "`%s' is a bad value for --duration", arg);
          else
            c->duration = val;
        }
      break;
    case CLIP_POSITION:
        {
          if (c->splice_in_at)
            argp_error (state, "can't use --position with --splice-in");
          else if (c->splice_over_at)
            argp_error (state, "can't use --position with --splice-over");
          double val = strtod (arg, &end);
          if (*end != '\0' || val < 0)
            argp_error (state, "`%s' is a bad value for --position", arg);
          else
            c->position = val;
        }
      break;
    case CLIP_RELATIVE_POSITION:
        {
          if (c->splice_in_at)
            argp_error (state, "can't use --relative-pos with --splice-in");
          else if (c->splice_over_at)
            argp_error (state, "can't use --relative-pos with --splice-over");
          double val = strtod (arg, &end);
          if (*end != '\0')
            argp_error (state, "`%s' is a bad value for --relative-position", arg);
          else
            c->relative_position = val;
        }
      break;
    case CLIP_VOLUME:
        {
          double val = strtod (arg, &end);
          if (*end != '\0' || val < 0 || val > 100)
            argp_error (state, "`%s' is a bad value for --volume", arg);
          else
            c->volume = val;
        }
      break;
    case CLIP_GREENSCREEN:
      c->greenscreen = 1;
      break;
    case CLIP_GREENSCREEN_THRESHOLD:
        {
          double val = strtod (arg, &end);
          if (*end != '\0' || val < 0 || val > 100)
            argp_error (state, "`%s' is a bad value for --greenscreen-threshold", arg);
          else
            c->greenscreen_threshold = val;
        }
      break;
    case CLIP_GREENSCREEN_DISTANCE:
        {
          double val = strtod (arg, &end);
          if (*end != '\0' || val < 0 || val > 100)
            argp_error (state, "`%s' is a bad value for --greenscreen-distance", arg);
          else
            c->greenscreen_distance = val;
        }
      break;
    case CLIP_TRANSITION_NAME:
      if (arg)
        {
          if (is_a_transition (arg))
            c->transition_name = arg;
          else
            {
              char *slash = strrchr (state->argv[0], '/');
              if (slash)
                slash++;
              else
                slash = state->argv[0];
              fprintf (stderr, "%s: `%s' is not in our list of transitions.  using anyway.\n", slash, arg);
              c->transition_name = arg;
            }
        }
      else
        c->transition_name = "crossfade";
      if (c->transition_overlap == 0)
        c->transition_overlap = 1.0;
      break;
    case CLIP_TRANSITION_OVERLAP:
        {
          if (arg)
            {
              double val = strtod (arg, &end);
              if (*end != '\0' || val < 0)
                argp_error (state, "`%s' is a bad value for --overlap", arg);
              else
                {
                  c->transition_overlap = val;
                  if (c->transition_name == NULL)
                    c->transition_name = "crossfade";
                }
            }
          else
            {
              c->transition_overlap = 1.0;
              if (c->transition_name == NULL)
                c->transition_name = "crossfade";
            }
        }
      break;
    case CLIP_X:
        {
          long int val = strtol (arg, &end, 0);
          if (*end != '\0')
            argp_error (state, "`%s' is a bad value for -x", arg);
          else
            c->posx = val;
        }
      break;
    case CLIP_Y:
        {
          long int val = strtol (arg, &end, 0);
          if (*end != '\0')
            argp_error (state, "`%s' is a bad value for -y", arg);
          else
            c->posy = val;
        }
      break;
    case CLIP_SCALE_WIDTH:
        {
          unsigned long int val = strtoul (arg, &end, 0);
          if (*end != '\0')
            argp_error (state, "`%s' is a bad value for --scale-width", arg);
          else
            c->width = val;
        }
      break;
    case CLIP_SCALE_HEIGHT:
        {
          unsigned long int val = strtoul (arg, &end, 0);
          if (*end != '\0')
            argp_error (state, "`%s' is a bad value for --scale-height", arg);
          else
            c->height = val;
        }
      break;
    case CLIP_CROP:
        {
          unsigned int a, b, d, e;
          int retval = sscanf (arg, "%u:%u:%u:%u", &a, &b, &d, &e);
          if (retval != 4)
            argp_error (state, "`%s' is a bad value for --crop", arg);
          else
            {
              c->crop = 1;
              c->crop_width = a;
              c->crop_height = b;
              c->crop_x = d;
              c->crop_y = e;
            }
        }
      break;
    case CLIP_FADE_IN:
        {
          if (arg)
            {
              double val = strtod (arg, &end);
              if (*end != '\0' || val < 0)
                argp_error (state, "`%s' is a bad value for --fade-in", arg);
              else
                c->fade_in = val;
            }
          else
            c->fade_in = 1.0;
        }
      break;
    case CLIP_FADE_OUT:
        {
          if (arg)
            {
              double val = strtod (arg, &end);
              if (*end != '\0' || val < 0)
                argp_error (state, "`%s' is a bad value for --fade-out", arg);
              else
                c->fade_out = val;
            }
          else
            c->fade_out = 1.0;
        }
      break;
    case CLIP_NO_VOLUME_FADE:
      c->no_volume_fade = 1;
      break;
    case CLIP_ZOOM:
      if (arg)
        {
          unsigned int a, b;
          double d;
          int retval = sscanf (arg, "%u:%u:%lf", &a, &b, &d);
          if (retval != 3 || d <= 0.0)
            argp_error (state, "`%s' is a bad value for --zoom", arg);
          else
            {
              c->zoom = 1;
              c->zoom_x = a;
              c->zoom_y = b;
              c->zoom_percent = d;
            }
        }
      else
        {
          c->zoom = 2; //means auto-center
          c->zoom_x = 0;
          c->zoom_y = 0;
          c->zoom_percent = 150.0;
        }
      break;
    case CLIP_IGNORE_AUDIO_TRACK:
      c->ignore_audio = 1;
      break;
    case CLIP_SPLICE_IN_AT:
        {
          if (c->splice_over_at)
            argp_error (state, "can't specify --splice-in with --splice-over");
          double val = strtod (arg, &end);
          if (*end != '\0' || val <= 0)
            argp_error (state, "`%s' is a bad value for --splice-in", arg);
          else
            c->splice_in_at = val;
        }
      break;
    case CLIP_SPLICE_OVER_AT:
        {
          if (c->splice_in_at)
            argp_error (state, "can't specify --splice-over with --splice-it");
          double val = strtod (arg, &end);
          if (*end != '\0' || val <= 0)
            argp_error (state, "`%s' is a bad value for --splice-ovet", arg);
          else
            c->splice_over_at = val;
        }
      break;
    }
  return 0;
}

static struct argp_option 
clip_options[] = 
{
    { 0, 0, 0, 0, "Clip Options:", CLIP_OPTION },
    { "start", CLIP_START, "SECS", 0,
      "The offset in the source FILE when the clip should begin playing (default 0)", CLIP_OPTION },
    { "duration", CLIP_DURATION, "SECS", 0,
      "How long the clip should play for (default is length of file)", CLIP_OPTION },
    { "position", CLIP_POSITION, "SECS", 0,
      "Where in the timeline the clip should be placed", CLIP_OPTION },
    { "relative-pos", CLIP_RELATIVE_POSITION, "SECS", 0,
      "Same as -p but relative to the end of the previous clip on this layer", CLIP_OPTION },
    { "volume", CLIP_VOLUME, "PERCENT", 0,
      "How loud the clip should be (default 100)", CLIP_OPTION },
    { "greenscreen", CLIP_GREENSCREEN, 0, 0,
      "Green regions become transparent so clips on subsequent layers show through", CLIP_OPTION },
    { "greenscreen-threshold", CLIP_GREENSCREEN_THRESHOLD, "PERCENT", OPTION_HIDDEN,
      "How closely we have to match the green (default 100)", CLIP_OPTION },
    { "greenscreen-distance", CLIP_GREENSCREEN_DISTANCE, "PERCENT", OPTION_HIDDEN,
      "How closely we have to match the green (default 28.7)", CLIP_OPTION },
    { "transition", CLIP_TRANSITION_NAME, "NAME", OPTION_ARG_OPTIONAL,
      "Animate from prior clip with this transition type", CLIP_OPTION+1},
    { "overlap", CLIP_TRANSITION_OVERLAP, "SECS", OPTION_ARG_OPTIONAL,
      "Transition from previous clip over SECS seconds", CLIP_OPTION+1},
    { NULL, CLIP_X, "PIXELS", 0,
      "Move the clip up/down on the video (default 0)", CLIP_OPTION},
    { NULL, CLIP_Y, "PIXELS", 0,
      "Move the clip right/left on the video (default 0)", CLIP_OPTION},
    { "scale-width", CLIP_SCALE_WIDTH, "PIXELS", 0,
      "Shrink/grow the clip width", CLIP_OPTION+1},
    { "scale-height", CLIP_SCALE_HEIGHT, "PIXELS", 0,
      "Shrink/grow the clip height", CLIP_OPTION+1},
    { "crop", CLIP_CROP, "W:H:X:Y", 0,
      "Only use a particular region of the video", CLIP_OPTION +1 },
    { "fade-in", CLIP_FADE_IN, "SECS", OPTION_ARG_OPTIONAL,
      "Increase opacity over the initial SECS seconds", CLIP_OPTION+1},
    { "fade-out", CLIP_FADE_OUT, "SECS", OPTION_ARG_OPTIONAL,
      "Decrease opacity over the final SECS seconds", CLIP_OPTION+1},
    { "no-volume-fade", CLIP_FADE_OUT, 0, OPTION_HIDDEN,
      "do not fade the volume in or out", CLIP_OPTION+1},
    { "scale", CLIP_ZOOM, "X:Y:PERCENT", OPTION_ARG_OPTIONAL,
      "Alias for -W, -H, -x, -y, but zooms in on a point", CLIP_OPTION+1},
    { "zoom", CLIP_ZOOM, "X:Y:PERCENT", OPTION_ALIAS | OPTION_HIDDEN,
      "Alias for -W, -H, -x, -y, but zooms in on a point", CLIP_OPTION+1},
    { "ignore-audio-track", CLIP_IGNORE_AUDIO_TRACK, 0, OPTION_HIDDEN,
      "Omit the audio track", CLIP_OPTION+1},
    { "splice-in", CLIP_SPLICE_IN_AT, "SECS", 0,
      "Interject this clip at SECS in the timeline", CLIP_OPTION+1},
    { "splice-over", CLIP_SPLICE_OVER_AT, "SECS", 0,
      "Like --splice-in, but underlying clips resume after as if they had been playing concurrently", CLIP_OPTION+1},
    { 0 }
};

int
is_a_clip_option (int key)
{
  switch (key)
    {
    case CLIP_START:
    case CLIP_DURATION:
    case CLIP_POSITION:
    case CLIP_RELATIVE_POSITION:
    case CLIP_GREENSCREEN:
    case CLIP_GREENSCREEN_THRESHOLD:
    case CLIP_GREENSCREEN_DISTANCE:
    case CLIP_VOLUME:
    case CLIP_TRANSITION_NAME:
    case CLIP_TRANSITION_OVERLAP:
    case CLIP_X:
    case CLIP_Y:
    case CLIP_SCALE_WIDTH:
    case CLIP_SCALE_HEIGHT:
    case CLIP_CROP:
    case CLIP_FADE_IN:
    case CLIP_FADE_OUT:
    case CLIP_NO_VOLUME_FADE:
    case CLIP_ZOOM:
    case CLIP_IGNORE_AUDIO_TRACK:
    case CLIP_SPLICE_IN_AT:
    case CLIP_SPLICE_OVER_AT:
      return 1;
    }
  return 0;
}

struct argp
clip_argp =
{
  clip_options, parse_clip_opt, 0, 0, 0
};

void
clip_validate (clip_t clip, struct argp_state *state)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  if (c->source_filename == NULL)
    argp_error (state, "each `clip' must have a --source");
}

int
clip_has_position (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->position != -1;
}

int
clip_has_relative_position (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->relative_position != 0.0;
}

double
clip_get_position (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->position;
}

void
clip_clear_position (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  c->position = -1;
  c->relative_position = 0;
}

double
clip_get_relative_position (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->relative_position;
}

int
clip_has_audio_track (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->cached_media_type & 0x02;
}
int
clip_has_video_track (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->cached_media_type & 0x04;
}

void
clip_remove_video_track (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  if (c->cached_media_type & 0x04)
    c->cached_media_type -=4;
}

void
clip_remove_audio_track (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  if (c->cached_media_type & 0x02)
    c->cached_media_type -=2;
}

void
clip_set_layer (clip_t clip, int layer)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  c->layer_number = layer;
}

int
clip_has_transition (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->transition_name || c->transition_overlap != 0.0;
}

char *
clip_get_uri (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->cached_uri;
}

int
clip_get_width (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->cached_width;
}

int
clip_get_height (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->cached_height;
}

int 
clip_has_crop (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->crop;
}

int
clip_crop_extents_in_bounds (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  unsigned int top = c->crop_y;
  unsigned int bottom = c->crop_y + c->crop_height;
  unsigned int left = c->crop_x;
  unsigned int right = c->crop_x + c->crop_width;
  if (top > c->cached_height)
    return 0;
  if (bottom > c->cached_height)
    return 0;
  if (left > c->cached_width)
    return 0;
  if (right > c->cached_width)
    return 0;
  return 1;
}

int
clip_has_fade (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->fade_in || c->fade_out;
}

int
clip_fade_in_bounds (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;

  unsigned long long int out;
  unsigned long long int in = c->start * gst_second;

  if (c->duration == 0)
    out = c->cached_duration - in;
  else
    {
      out = c->start + c->duration * gst_second;
      if (out > c->cached_duration)
        out = c->cached_duration;
    }
  unsigned long long int duration = out - in;
  if (c->fade_in * gst_second > duration)
    return 0;
  if (c->fade_out * gst_second > duration)
    return 0;
  return 1;
}

void
clip_set_file (clip_t clip, char *file)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  c->source_filename = file;
}

int
clip_has_zoom (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->zoom != 0;
}

int
clip_zoom_point_in_bounds (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  if (c->zoom_x >= c->cached_width)
    return 0;
  if (c->zoom_y >= c->cached_height)
    return 0;
  return 1;
}

void
clip_fill_zoom (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  if (c->zoom == 2)
    {
      c->zoom_x = c->cached_width / 2;
      c->zoom_y = c->cached_height / 2;
    }
  double p = c->zoom_percent / 100.0;
  c->width = c->cached_width * p;
  c->height = c->cached_height * p;
  c->posx = -((c->width - c->cached_width) / 2) + 
    (((c->cached_width/2)-c->zoom_x) *p);
  c->posy = -((c->height - c->cached_height) / 2) + 
    (((c->cached_height/2)-c->zoom_y)*p);
}

int 
clip_has_splice (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->splice_in_at || c->splice_over_at;
}

unsigned long long int
clip_get_cached_duration (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->cached_duration;
}

double
clip_get_duration (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->duration;
}

double
clip_get_start (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->start;
}

void
clip_set_start (clip_t clip, double start)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  c->start = start;
}

void
clip_set_duration (clip_t clip, double duration)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  c->duration = duration;
}

static void
fill_vars (struct clip_details_t *c, unsigned long long int *in, unsigned long long int *out, unsigned long long int *at, unsigned long long int *current_position)
{
  unsigned long long int _in, _out, _at, _current_position;
  _in = c->start * gst_second;
  if (c->duration == 0)
    _out = c->cached_duration;
  else
    {
      _out = (c->start + c->duration) * gst_second;
      if (_out > c->cached_duration)
        _out = c->cached_duration;
    }

  if (c->relative_position)
    {
      if (c->relative_position < 0)
        {
          double pos = *current_position / (double)gst_second;
          if (c->relative_position * -1 > pos)
            _at = 0;
          else
            _at = *current_position + (c->relative_position * gst_second);
        }
      else
        _at = *current_position + (c->relative_position * gst_second);
    }
  else
    {
      if (c->position >= 0)
        _at = c->position * gst_second;
      else
        _at = *current_position;
    }
  if (c->transition_overlap)
    _at -= c->transition_overlap * gst_second;

  *current_position = _at + _out - _in;
  c->cached_in = _in;
  c->cached_out = _out;
  c->cached_at = _at;
  if (in)
    *in = _in;
  if (out)
    *out = _out;
  if (at)
    *at = _at;
}

static unsigned long long int
clip_is_at (struct clip_details_t *clip, unsigned long long int *duration, clip_t *clips, int num_clips)
{
  unsigned long long int current_position = 0;
  int i;
  unsigned long long int at;
  for (i = 0; i < num_clips; i++)
    {
      struct clip_details_t *c = clips[i];
      unsigned long long int in, out;
      if (c->layer_number == clip->layer_number && !clip_has_splice (clips[i]) &&
          clip_has_video_track (clips[i]))
        fill_vars (c, &in, &out, &at, &current_position);

      if (clip == c)
        {
          if (in > out)
            {
              printf("crappers %lld > %lld\n", in, out);
              exit (0);
            }
          if (duration)
            *duration = out - in;
          break;
        }
    }
  return at;
}

static int
clip_is_on_layer (clip_t clip, int layer)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->layer_number == layer;
}

static int
which_clip_is_at (double secs, int layer, unsigned long long int *position, clip_t *clips, int num_clips)
{
  int i;
  for (i = 0; i < num_clips; i++)
    {
      if (clip_is_on_layer (clips[i], layer) && !clip_has_splice (clips[i]) &&
          clip_has_video_track (clips[i]))
        {
          unsigned long long duration;
          unsigned long long int at = 
            clip_is_at (clips[i], &duration, clips, num_clips);
          unsigned long long int target = secs * gst_second;
          if (target >= at && target <= at + duration)
            {
              *position = at;
              return i;
            }
        }
    }

  return -1;
}

static int
find_idx_of_clip (clip_t clip, clip_t *clips, int num_clips)
{
  int i;
  for (i = 0; i < num_clips; i++)
    if (clips[i] == clip)
      return i;
  return -1;
}

static int
find_idx_of_clip_before (double target, int layer, int audio, clip_t *clips, int num_clips)
{
  int i;
  int idx = -1;
  unsigned long long int current_position = 0;
  for (i = 0; i < num_clips; i++)
    {
      struct clip_details_t *c = clips[i];
      if (c->layer_number == layer)
        {
          if ((!audio && clip_has_video_track (c) ) ||
              (audio && clip_has_audio_track (c)))
            {
              unsigned long long int in, out, at;
              fill_vars (c, &in, &out, &at, &current_position);
              if (at + (out - in) > target * gst_second)
                break;
              idx = i;
            }
        }
    }
  return idx;
}

static int
clip_splice_in  (struct clip_details_t *c, clip_t **clips, int *num_clips)
{
  unsigned long long int pos;
  int idx = which_clip_is_at (c->splice_in_at, c->layer_number, &pos, 
                              *clips, *num_clips);
  if (idx != -1)
    {
      //okay now we do the splice.
      //c is the clip we want to add.
      //clips[idx] is the clip we want to cut in two pieces
      //
      //we shorten clips[idx], and then add another clip after c.
      unsigned long long int at = c->splice_in_at * gst_second;
      double old_duration = clip_get_duration ((*clips)[idx]);
      if (old_duration == 0)
        old_duration = (clip_get_cached_duration ((*clips)[idx]) / gst_second) - clip_get_start ((*clips)[idx]);
      double new_duration = (double)(at - pos) / gst_second;
      clip_set_duration ((*clips)[idx], new_duration);
      clip_t newclip;
      init_clip_from_clip (&newclip, (*clips)[idx]);
      clip_clear_position (newclip);
      clip_set_start (newclip, clip_get_start ((*clips)[idx]) + new_duration);

      clip_set_duration (newclip, old_duration - new_duration);
      int nidx = find_idx_of_clip (c, *clips, *num_clips);
      insert_clip_after (newclip, nidx, clips, num_clips);
    }
  else
    {
      //what do we do?
      //well, we just put it at that spot in the timeline.
      c->position = c->splice_in_at;
    }
  c->splice_in_at = 0;
  return 0;
}

static int
remove_clip (clip_t clip, int retain_audio, clip_t **clips, int *num_clips)
{
  int removed = 0;
  int i;
  for (i = 0; i < *num_clips; i++)
    {
      if ((*clips)[i] == clip)
        {
          if (retain_audio)
            clip_remove_video_track ((*clips)[i]);
          else
            {
              int j;
              destroy_clip (clip);
              for (j = i+1; j <= *num_clips - 1; j++)
                (*clips)[j-1] = (*clips)[j];
              (*num_clips)--;
              //not sure if this is necessary, when resizing smaller
              //*clips = realloc (*clips, sizeof (clip_t) * *num_clips);
            }
          removed = 1;
          break;
        }
    }
  return removed;
}

struct overlap_t
{
  clip_t clip;
  int overlap_start;
  int overlap_end;
  int wholly_overlapped;
};
static void
dump_overlap (struct overlap_t *overlap, int num_overlap)
{
  int i;
  for (i = 0; i < num_overlap; i++)
    {
      printf("%d. %d %d %d\n", i, overlap[i].overlap_start,overlap[i].overlap_end,overlap[i].wholly_overlapped);
    }
}

static void
which_clips_overlap (double secs, double duration, int layer, unsigned long long int *start_pos, unsigned long long int *end_pos, clip_t *clips, int num_clips, struct overlap_t **overlap, int *num_overlap)
{
  int i;
  for (i = 0; i < num_clips; i++)
    {
      if (clip_is_on_layer (clips[i], layer) && !clip_has_splice (clips[i]))
        {
          unsigned long long clip_duration;
          unsigned long long int at = 
            clip_is_at (clips[i], &clip_duration, clips, num_clips);
          unsigned long long int starttarget = secs * gst_second;
          unsigned long long int endtarget = (secs+duration) * gst_second;
          //is the left edge anywhere in the range?
          //is the right edge anywhere in the range?
          //are both edges outside, encompassing the range?
          if ((at >= starttarget && at <= endtarget) ||
              (at+clip_duration >= starttarget && at+clip_duration <= endtarget) ||
              (at < starttarget && at+clip_duration>endtarget))
            {
              if (*start_pos == 0)
                *start_pos = at;
              *end_pos = at;
              (*num_overlap)++;
              *overlap = realloc (*overlap, sizeof (struct overlap_t) * *num_overlap);
              struct overlap_t item;
              memset (&item, 0, sizeof (struct overlap_t));
              item.clip = clips[i];
              item.overlap_start = (at <= starttarget && at+clip_duration >= starttarget);
              item.overlap_end = (at <= endtarget && at+clip_duration >= endtarget);
              item.wholly_overlapped = !item.overlap_start && !item.overlap_end;
              if (!item.wholly_overlapped)
                {
                  if (item.overlap_start && item.overlap_end)
                    item.wholly_overlapped = 1;
                }
              if (!item.wholly_overlapped)
                item.wholly_overlapped = (at >= starttarget && at+clip_duration <= endtarget);
              (*overlap)[(*num_overlap)-1] = item;
              //hey we found the first clip that overlaps.
              //how many more do?
            }
        }
    }

  return;
}

void
clip_clear_fade (clip_t clip)
{
  struct clip_details_t *c = clip;
  c->fade_in = 0;
  c->fade_out = 0;
}

void
clip_clear_transition (clip_t clip)
{
  struct clip_details_t *c = clip;
  c->transition_name = NULL;
  c->transition_overlap = 0;
}

void
clip_set_fade_in (clip_t clip, double fade)
{
  struct clip_details_t *c = clip;
  c->fade_in = fade;
}
void
clip_set_fade_out (clip_t clip, double fade)
{
  struct clip_details_t *c = clip;
  c->fade_out = fade;
}

static int
clip_splice_over (struct clip_details_t *c, clip_t **clips, int *num_clips)
{
  int i;
  unsigned long long int start_pos = 0, end_pos = 0;
  double duration = c->duration;
  if (duration == 0)
    duration = c->cached_duration;
  struct overlap_t *overlap = NULL;
  int num_overlap = 0;
  which_clips_overlap (c->splice_over_at, duration, c->layer_number, 
                       &start_pos, &end_pos, *clips, *num_clips,
                       &overlap, &num_overlap);
  if (!num_overlap)
    {
      c->position = c->splice_over_at;
      c->splice_over_at = 0;
      //we didn't hit anything?  what do we do?
      return 0;
    }

  clip_t start = NULL;
  //handle start overlap:
  for (i = 0; i < num_overlap; i++)
    {
      if (overlap[i].overlap_start)
        {
          start = overlap[i].clip;
          break;
        }
    }

  clip_t end = NULL;
  //handle end overlap:
  for (i = 0; i < num_overlap; i++)
    {
      if (overlap[i].overlap_end)
        end = overlap[i].clip;
    }
  int idx = find_idx_of_clip (c, *clips, *num_clips);
  if (!start)
    c->position = c->splice_over_at;
  if (start == end)
    {
      //hey we have a special case here, where we have to make a new clip.
      //this is where our clip START completely subsumes clip C.
              
      int startidx = find_idx_of_clip (start, *clips, *num_clips);
      if (idx != startidx + 1)
        {
          clip_t copy = NULL;
          init_clip_from_clip (&copy, c);
          clip_clear_position (copy);
          remove_clip (c, 0, clips, num_clips);
          insert_clip_after (copy,  startidx, clips, num_clips);
          c = (struct clip_details_t*) copy;
          idx = startidx + 1;
        }
      double old_splice_over_at = c->splice_over_at;
      c->splice_over_at = 0;
      unsigned long long int d;
      unsigned long long int at = clip_is_at (c, &d, *clips, *num_clips);
      unsigned long long int startat = clip_is_at (start, &d, *clips, *num_clips);

      double cduration = c->duration;
      if (cduration == 0)
        cduration = (double)c->cached_duration / gst_second;
      clip_t newend = NULL;
      init_clip_from_clip (&newend, start);
      clip_clear_position (newend);
      clip_set_fade_in (newend, 0);
      clip_set_fade_out (start, 0);
      clip_clear_transition (newend);
      clip_set_duration (newend,
                         (((double)(startat+d)/gst_second) - ((old_splice_over_at+cduration))));

      insert_clip_after (newend, idx, clips, num_clips);
      double ss = 0;
      if (startat > 0)
        ss = (double)startat / gst_second;
      clip_set_duration (start, old_splice_over_at-ss);

      clip_set_start (newend, clip_get_start (newend) + ((old_splice_over_at + cduration) - ss));
    }
  else
    {
      unsigned long long int ed, d;
      unsigned long long endat = 0, startat = 0;
      if (end)
        endat = clip_is_at (end, &ed, *clips, *num_clips);
      if (start)
        startat = clip_is_at (start, &d, *clips, *num_clips);
      if (start)
        {
          //the start of clip C overlaps clip START.
          //change the duration of clip START to be shortened
          clip_set_duration (start, c->splice_over_at - ((double)startat/gst_second));

          int startidx = find_idx_of_clip (start, *clips, *num_clips);
          if (idx != startidx + 1)
            {
              clip_t copy = NULL;
              init_clip_from_clip (&copy, c);
              clip_clear_position (copy);
              remove_clip (c, 0, clips, num_clips);
              insert_clip_after (copy,  startidx, clips, num_clips);
              c = (struct clip_details_t*) copy;
              idx = startidx + 1;
            }
        }
  if (num_overlap > 2)
    {
      //get rid of the ones that overlap completely
      for (i = 1; i < num_overlap - 1; i++)
        if (overlap[i].wholly_overlapped)
          remove_clip (overlap[i].clip, 0, clips, num_clips);
    }
      if (end)
        {
          //the end of our new clip overlaps this clip.
          //change the duration of clip END to be shortened
          //change the start of the clip END to be farther along
          //it will be placed at the right place in the timeline by virtue of the position of our clip C in the list of clips.

          int endidx = find_idx_of_clip (end, *clips, *num_clips);
          if (idx != endidx - 1)
            {
              clip_t copy = NULL;
              init_clip_from_clip (&copy, c);
              clip_clear_position (copy);
              remove_clip (c, 0, clips, num_clips);
              insert_clip_after (copy,  endidx-1, clips, num_clips);
              c = (struct clip_details_t*) copy;
              idx = endidx;
            }
          double old_splice_over_at = c->splice_over_at;
          c->splice_over_at = 0;

          double cduration = c->duration;
          if (cduration == 0)
            cduration = (double)c->cached_duration / gst_second;
          unsigned long long int d;
          unsigned long long int at = clip_is_at (c, &d, *clips, *num_clips);
          double old_duration = clip_get_duration (end);

          clip_set_duration (end, ((double)(endat + ed) / gst_second) - 
                             (old_splice_over_at + cduration));
          clip_set_start (end, clip_get_start (end) + 
                          old_duration - clip_get_duration (end));
        }
    }
    
  free (overlap);

  c->splice_over_at = 0;

  return 0;
}

double
clip_get_fade_in (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->fade_in;
}

double
clip_get_fade_out (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->fade_out;
}

int
clip_get_crop_width (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->crop_width;
}

int
clip_get_crop_height (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->crop_height;
}

int
clip_get_crop_x (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->crop_x;
}

int
clip_get_crop_y (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->crop_y;
}

int
clip_get_zoom_x (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->zoom_x;
}

int
clip_get_zoom_y (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->zoom_y;
}

double
clip_get_zoom_percent (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->zoom_percent;
}

int
clip_get_greenscreen (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->greenscreen;
}

char *
clip_get_transition_name (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->transition_name;
}

double
clip_get_transition_duration (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->transition_overlap;
}

int
clip_splice (clip_t clip, clip_t **clips, int *num_clips)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  if (c->splice_in_at)
    return clip_splice_in (c, clips, num_clips);
  else if (c->splice_over_at)
    return clip_splice_over (c, clips, num_clips);
  return 0;
}
    
int
clip_get_layer_number (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->layer_number;
}

/*
unsigned long long int
clip_get_position (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->cached_at;
}
*/

void
clip_set_position (clip_t clip, double pos)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  c->position = pos;
}

int
clip_is_audio_only (clip_t clip)
{
  struct clip_details_t *c = (struct clip_details_t*) clip;
  return c->cached_media_type & 0x02 && !(c->cached_media_type & 0x04);
}

int
make_hole_on_layer (clip_t **clips, int *num_clips, int layer, double at, double duration)
{
  int i;
  unsigned long long int start_pos = 0, end_pos = 0;
  struct overlap_t *overlap = NULL;
  int num_overlap = 0;
  which_clips_overlap (at, duration, layer, 
                       &start_pos, &end_pos, *clips, *num_clips,
                       &overlap, &num_overlap);
  if (!num_overlap)
    return 0;
  else if (num_overlap > 2)
    {
      for (i = 0; i < num_overlap; i++)
        if (overlap[i].wholly_overlapped)
          remove_clip (overlap[i].clip, 1, clips, num_clips);
    }

  clip_t start = NULL;
  //handle start overlap:
  for (i = 0; i < num_overlap; i++)
    {
      if (overlap[i].overlap_start)
        {
          start = overlap[i].clip;
          break;
        }
    }


  clip_t end = NULL;
  //handle end overlap:
  for (i = 0; i < num_overlap; i++)
    {
      if (overlap[i].overlap_end)
        end = overlap[i].clip;
    }

  if (start == end)
    {
      unsigned long long int d;
      unsigned long long int startat = clip_is_at (start, &d, *clips, *num_clips);
      clip_t middle = NULL;
      init_clip_from_clip (&middle, start);
      clip_set_duration (middle, duration);
      clip_clear_fade (middle);
      clip_clear_transition (middle);
      clip_remove_video_track (middle);
      clip_set_start (middle, clip_get_start (start) + (at - ((double)startat/gst_second)));
      clip_clear_position (middle);
      clip_set_position (middle, at);

      clip_t newend = NULL;
      init_clip_from_clip (&newend, start);
      clip_clear_position (newend);
      clip_set_fade_in (newend, 0);
      clip_set_fade_out (start, 0);
      clip_clear_transition (newend);
      clip_set_duration (newend,
                         (((double)(startat+d)/gst_second) - ((at+duration))));

      int idx = find_idx_of_clip (start, *clips, *num_clips);
      insert_clip_after (newend, idx, clips, num_clips);
      double ss = 0;
      if (startat > 0)
        ss = (double)startat / gst_second;
      clip_set_duration (start, at-ss);
      clip_clear_position (newend);
      clip_set_position (newend, at + duration);

      clip_set_start (newend, clip_get_start (newend) + ((at + duration) - ss));
      if (clip_has_audio_track (middle))
        insert_clip_after (middle, idx, clips, num_clips);
      else
        destroy_clip (middle);
    }
  else
    {
      //benfix add middle audio clip here too.
      unsigned long long int ed, d;
      unsigned long long endat, startat;
      if (end)
        endat = clip_is_at (end, &ed, *clips, *num_clips);
      if (start)
        startat = clip_is_at (start, &d, *clips, *num_clips);

      if (start)
        {
          double old_duration = clip_get_duration (start);
          clip_set_duration (start, at - ((double)startat/gst_second));
          clip_t middle = NULL;
          init_clip_from_clip (&middle, start);
          clip_set_duration (middle, old_duration - clip_get_duration (start));
          clip_clear_fade (middle);
          clip_clear_transition (middle);
          clip_remove_video_track (middle);
          clip_set_start (middle, clip_get_start (start) + (at - ((double)startat/gst_second)));
          clip_clear_position (middle);
          clip_set_position (middle, at);
          if (clip_has_audio_track (middle))
            {
              int idx = find_idx_of_clip (start, *clips, *num_clips);
              insert_clip_after (middle, idx, clips, num_clips);
            }
          else
            destroy_clip (middle);
        }
      if (end)
        {
          double old_duration = clip_get_duration (end);
          double old_start = clip_get_start (end);
          clip_set_duration (end, clip_get_duration (end) - 
                            ((at+duration) - ((double)endat/gst_second)));
          clip_set_start (end, clip_get_start (end) + 
                          old_duration - clip_get_duration (end));
          clip_clear_position (end);
          clip_set_position (end, at + duration);
          clip_t middle = NULL;
          init_clip_from_clip (&middle, end);
          clip_set_duration (middle, (at+duration) - ((double)endat/gst_second));
          clip_clear_fade (middle);
          clip_clear_transition (middle);
          clip_remove_video_track (middle);
          clip_set_start (middle, old_start);
          clip_clear_position (middle);
          //clip_set_position (middle, (double)endat/gst_second);
          if (clip_has_audio_track (middle))
            {
              int idx = find_idx_of_clip_before (at+duration, layer, 1, *clips, *num_clips);
              insert_clip_after (middle, idx, clips, num_clips);
            }
          else
            destroy_clip (middle);
        }
    }

  free (overlap);
  return 0;
}

static int
is_layer_contiguous_prior_to_point (clip_t *clips, int num_clips, int layer, double target)
{
  int i;
  unsigned long long int current_position = 0;
  unsigned long long int end;
  for (i = 0; i < num_clips; i++)
    {
      if (clip_is_on_layer (clips[i], layer) && clip_has_video_track (clips[i]))
        {
          unsigned long long int in, out, at;
          fill_vars (clips[i], &in, &out, &at, &current_position);
          if ((double)current_position/gst_second > target)
            break;
          else if ((double)current_position/gst_second == target)
            {
              end = at + (out - in);
              break;
            }
        }
    }
  return (double)end/gst_second == target;
}

static void
move_clips_to_layer (double at, double duration, int layer, int dest_layer,
                     clip_t **clips, int *num_clips)
{
  //okay, now we make our cuts and promote one or more clips to layer 0
  int i;
  unsigned long long int start_pos = 0, end_pos = 0;
  struct overlap_t *overlap = NULL;
  int num_overlap = 0, count = 0;
  int idx = find_idx_of_clip_before (at, dest_layer, 0, *clips, *num_clips);

  which_clips_overlap (at, duration, layer, &start_pos, &end_pos, 
                       *clips, *num_clips, &overlap, &num_overlap);
  for (i = 0; i < num_overlap; i++)
    {
      if (overlap[i].wholly_overlapped)
        {
          if (count == 0 && !is_layer_contiguous_prior_to_point (*clips, *num_clips, 
                                                             dest_layer, at))
            {
              unsigned long long int at = clip_is_at (overlap[i].clip, NULL, 
                                                      *clips, *num_clips);
              clip_clear_position (overlap[i].clip);
              clip_set_position (overlap[i].clip, (double)at/gst_second);
            }
          count++;
          clip_set_layer (overlap[i].clip, dest_layer);
        }
    }
  //now we move the clips around in the clips array

  //here we find where to put them

  //here we insert copies of them there
  for (i = 0; i < num_overlap; i++)
    if (overlap[i].wholly_overlapped)
      {
        clip_t copy = NULL;
        init_clip_from_clip (&copy, overlap[i].clip);
        clip_remove_audio_track (copy);
        insert_clip_after (copy, idx, clips, num_clips);
        idx++;
      }

  //and then we remove the originals from elsewhere
  for (i = 0; i < num_overlap; i++)
    if (overlap[i].wholly_overlapped)
      remove_clip (overlap[i].clip, 0, clips, num_clips);
  free (overlap);
}

int
clip_promote (promote_t promote, clip_t **clips, int *num_clips, int num_layers)
{
  struct promote_details_t *p = promote;
  int i;
  //make a hole at p->start until p->duration on all layers except for the one 
  //we want to promote
  for (i = 0; i < num_layers; i++)
    {
      if (promote_get_layer (promote) == i)
        continue;
      make_hole_on_layer (clips, num_clips, i, promote_get_position (p), 
                          promote_get_duration (p));
    }
  if (promote_get_justmakehole (p))
    return 0;

  //okay, we need to make maybe two cuts:
  //1. at position
  //2. at position+duration
  //these cuts can happen on the same clip, or two separate clips.
  unsigned long long int startat, endat;
  int startidx = which_clip_is_at (promote_get_position (p), promote_get_layer (p), &startat, *clips, *num_clips);
  int endidx = which_clip_is_at (promote_get_position (p) + promote_get_duration (p), promote_get_layer (p), &endat, *clips, *num_clips);
  clip_t start = NULL;
  if (startidx > -1)
    start = (*clips)[startidx];
  clip_t end = NULL;
  if (endidx > -1)
    end = (*clips)[endidx];
  double at = promote_get_position (promote);
  double duration = promote_get_duration (promote);
  if (start == end)
    {
      unsigned long long int d;
      unsigned long long int startat = clip_is_at (start, &d, *clips, *num_clips);
      clip_t middle= NULL;
      init_clip_from_clip (&middle, start);
      clip_clear_fade (middle);
      clip_clear_position (middle);
      clip_clear_transition (middle);
      clip_set_duration (middle, promote_get_duration (p));
      clip_set_start (middle, clip_get_start (middle) + (at-((double)startat/gst_second)));

      clip_t newend = NULL;
      init_clip_from_clip (&newend, start);
      clip_set_fade_in (newend, 0);
      clip_set_fade_out (start, 0);
      clip_clear_position (newend);
      clip_clear_transition (newend);
      clip_set_duration (newend,
                         (((double)(startat+d)/gst_second) - ((at+duration))));

      int idx = find_idx_of_clip (start, *clips, *num_clips);
      insert_clip_after (middle, idx, clips, num_clips);
      idx++;
      insert_clip_after (newend, idx, clips, num_clips);
      double ss = 0;
      if (startat > 0)
        ss = (double)startat / gst_second;
      clip_set_duration (start, at-ss);
      clip_set_position (newend, at + duration);

      clip_set_start (newend, clip_get_start (newend) + ((at + duration) - ss));
    }
  else
    {
          unsigned long long int ed;
          unsigned long long endat = clip_is_at (end, &ed, *clips, *num_clips);
      if (start)
        {
          clip_t middle = NULL;
          init_clip_from_clip (&middle, start);
          clip_clear_position (middle);
          clip_clear_fade (middle);
          clip_clear_transition (middle);

          unsigned long long int d;
          unsigned long long int startat = clip_is_at (start, &d, *clips, *num_clips);
          clip_set_duration (middle, ((double)(startat+d)/gst_second)-at);
          clip_set_start (middle, clip_get_start (start) + clip_get_duration (start) - clip_get_duration (middle));
          int idx = find_idx_of_clip (start, *clips, *num_clips);
          insert_clip_after (middle, idx, clips, num_clips);

          clip_set_duration (start, at - ((double)startat/gst_second));

        }
      if (end)
        {
          clip_t middle = NULL;
          init_clip_from_clip (&middle, end);
          clip_clear_position (end);
          clip_clear_position (middle);
          clip_clear_fade (middle);
          clip_clear_transition (middle);
          int idx = find_idx_of_clip_before ((double)endat/gst_second, promote_get_layer (p), 0, *clips, *num_clips);
          clip_set_duration (middle, (at+duration)-((double)endat/gst_second));
          clip_set_start (middle, clip_get_start (end));
          insert_clip_after (middle, idx, clips, num_clips);
          double old_duration = clip_get_duration (end);
          clip_set_position (end, at + duration);
          clip_set_duration (end, ((double)(endat + ed) / gst_second) - 
                             (at + duration));
          clip_set_start (end, clip_get_start (end) + clip_get_duration (middle));
        }
    }

  //destination layer will be zero by default
  move_clips_to_layer (at, duration, promote_get_layer (p), 
                       promote_get_destination_layer (p), clips, num_clips);
  return 0;
}
