// Copyright (C) 2017 Ben Asselstine
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <limits.h>
#include <stdio.h>
#include <error.h>
#include <argz.h>
#include <string.h> //remove me
#include <time.h>
#include <stdlib.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk-pixbuf/gdk-pixbuf-loader.h>
#include <glib.h>
#include "generate.h"
#include "clip-priv.h"
#include <sys/time.h>
#include "layer-priv.h"
#include <gst/gst.h>
#include <gst/pbutils/pbutils.h>

static char *
get_media_uri (char *filename)
{
  GFile *f = g_file_new_for_commandline_arg (filename);
  char *uri = strdup (g_file_get_uri (f));
  g_object_unref (f);
  return uri;
}

static int
is_image (char *filename)
{
  GError *err = NULL;
  GdkPixbuf * pixbuf = gdk_pixbuf_new_from_file (filename, &err);
  int retval = pixbuf != NULL;
  if (pixbuf)
    g_object_unref (pixbuf);
  return retval;
}

typedef struct _CustomData 
{
  GstDiscoverer *discoverer;
  GMainLoop *loop;
  unsigned long long int duration;
  int mediatype;
  int width;
  int height;
} CustomData;

static void
has_audio_video (GstDiscovererStreamInfo *info, int *has_audio, int *has_video)
{
  if (!info)
    return;

  if (strcmp (gst_discoverer_stream_info_get_stream_type_nick (info), "video") == 0)
    *has_video = 1;
  else if (strcmp (gst_discoverer_stream_info_get_stream_type_nick (info), "audio") == 0)
    *has_audio = 1;

  GstDiscovererStreamInfo *next = gst_discoverer_stream_info_get_next (info);
  if (next) 
    {
      has_audio_video (next, has_audio, has_video);
      gst_discoverer_stream_info_unref (next);
    } 
  else if (GST_IS_DISCOVERER_CONTAINER_INFO (info)) 
    {
      GList *tmp, *streams;

      streams = gst_discoverer_container_info_get_streams (GST_DISCOVERER_CONTAINER_INFO (info));
      for (tmp = streams; tmp; tmp = tmp->next) 
        {
          GstDiscovererStreamInfo *tmpinf = (GstDiscovererStreamInfo *) tmp->data;
          has_audio_video (tmpinf, has_audio, has_video);
        }
      gst_discoverer_stream_info_list_free (streams);
    }
}

static void on_mediatype_cb (GstDiscoverer *discoverer, GstDiscovererInfo *info, GError *err, CustomData *data) 
{
  GstDiscovererResult result = gst_discoverer_info_get_result (info);
  GstDiscovererStreamInfo *sinfo;
  if (result == GST_DISCOVERER_ERROR)
    {
      fprintf (stderr, "%s\n", err->message);
      return;
    }
  if (result != GST_DISCOVERER_OK)
    return;
  sinfo = gst_discoverer_info_get_stream_info (info);
  if (!sinfo)
    return;
  int has_audio = 0, has_video = 0;
  has_audio_video (sinfo, &has_audio, &has_video);
  gst_discoverer_stream_info_unref (sinfo);
  if (has_audio)
    data->mediatype += 2;
  if (has_video)
    data->mediatype += 4;
  return;
}

static void on_duration_cb (GstDiscoverer *discoverer, GstDiscovererInfo *info, GError *err, CustomData *data) 
{
  GstDiscovererResult result = gst_discoverer_info_get_result (info);
  if (result == GST_DISCOVERER_ERROR)
    {
      fprintf (stderr, "%s\n", err->message);
      return;
    }
  if (result != GST_DISCOVERER_OK)
    return;
  data->duration = gst_discoverer_info_get_duration (info);
}

static void on_dimensions_cb (GstDiscoverer *discoverer, GstDiscovererInfo *info, GError *err, CustomData *data) 
{
  GstDiscovererResult result = gst_discoverer_info_get_result (info);
  if (result == GST_DISCOVERER_ERROR)
    {
      fprintf (stderr, "%s\n", err->message);
      return;
    }
  if (result != GST_DISCOVERER_OK)
    return;
  GList *list = gst_discoverer_info_get_video_streams (info);
  if (list)
    {
      if (g_list_length (list) > 0)
        {
          data->width = 
            gst_discoverer_video_info_get_width (GST_DISCOVERER_VIDEO_INFO (list->data));
          data->height = 
            gst_discoverer_video_info_get_height (GST_DISCOVERER_VIDEO_INFO (list->data));
        }
      gst_discoverer_stream_info_list_free (list);
    }
  return;
}

static void on_finished_cb (GstDiscoverer *discoverer, CustomData *data) 
{
  g_main_loop_quit (data->loop);
}
   
static unsigned long long int
get_media_duration (char *uri)
{
  GError *err = NULL;
  CustomData data;
  memset (&data, 0, sizeof (data));
  data.discoverer = gst_discoverer_new (5 * GST_SECOND, &err);
  if (!data.discoverer)
    return 0;
  g_signal_connect (data.discoverer, "discovered", G_CALLBACK (on_duration_cb), &data);
  g_signal_connect (data.discoverer, "finished", G_CALLBACK (on_finished_cb), &data);
  gst_discoverer_start (data.discoverer);
  if (!gst_discoverer_discover_uri_async (data.discoverer, uri)) 
    return 0;
  data.loop = g_main_loop_new (NULL, FALSE);
  g_main_loop_run (data.loop);
  gst_discoverer_stop (data.discoverer);
  g_object_unref (data.discoverer);
  g_main_loop_unref (data.loop);

  return data.duration;
}

static int
get_media_type (char *uri)
{
  GError *err = NULL;
  CustomData data;
  memset (&data, 0, sizeof (data));
  data.discoverer = gst_discoverer_new (5 * GST_SECOND, &err);
  if (!data.discoverer)
    return 0;
  g_signal_connect (data.discoverer, "discovered", G_CALLBACK (on_mediatype_cb), &data);
  g_signal_connect (data.discoverer, "finished", G_CALLBACK (on_finished_cb), &data);
  gst_discoverer_start (data.discoverer);
  if (!gst_discoverer_discover_uri_async (data.discoverer, uri)) 
    return 0;
  data.loop = g_main_loop_new (NULL, FALSE);
  g_main_loop_run (data.loop);
  gst_discoverer_stop (data.discoverer);
  g_object_unref (data.discoverer);
  g_main_loop_unref (data.loop);
  return data.mediatype;
}

static int
get_media_dimensions (char *uri, int *width, int *height)
{
  GError *err = NULL;
  CustomData data;
  memset (&data, 0, sizeof (data));
  data.discoverer = gst_discoverer_new (5 * GST_SECOND, &err);
  if (!data.discoverer)
    return 0;
  g_signal_connect (data.discoverer, "discovered", G_CALLBACK (on_dimensions_cb), &data);
  g_signal_connect (data.discoverer, "finished", G_CALLBACK (on_finished_cb), &data);
  gst_discoverer_start (data.discoverer);
  if (!gst_discoverer_discover_uri_async (data.discoverer, uri)) 
    return 0;
  data.loop = g_main_loop_new (NULL, FALSE);
  g_main_loop_run (data.loop);
  gst_discoverer_stop (data.discoverer);
  g_object_unref (data.discoverer);
  g_main_loop_unref (data.loop);

  *width = data.width;
  *height = data.height;
  return 1;
}

static void
dump_footer (FILE *fp)
{
  fprintf (fp, "</project>\n");
  fprintf (fp, "</ges>\n");
}

static char *
munge_commandline (char *commandline)
{
  char *letter = NULL;
  char *c = g_markup_escape_text (commandline, -1);
  char *s = NULL;
  size_t len = 1;
  s = calloc (1, sizeof (char) * len);
  for (letter = c; *letter != '\0'; letter++)
    {
      if (strncmp (letter, CLIP_CMD, strlen (CLIP_CMD)) == 0 ||
          strncmp (letter, LAYER_CMD, strlen (LAYER_CMD)) == 0 ||
          strncmp (letter, PROMOTE_CMD, strlen (PROMOTE_CMD)) == 0)
        {
          len += 5;
          s = realloc (s, sizeof (char) * len);
          s[len-6]='\\';
          s[len-5]='\n';
          s[len-4]=' ';
          s[len-3]=' ';
          s[len-2]='+';
          s[len-1] = '\0';
        }
      else if (strncmp (letter, "--", 2) == 0)
        {
          len +=3;
          s = realloc (s, sizeof (char) * len);
          s[len-4] = *letter;
          s[len-3] = '\\';
          s[len-2] = *letter;
          s[len-1] = '\0';
          letter++;
        }
      else
        {
          len++;
          s = realloc (s, sizeof (char) * len);
          s[len-2] = *letter;
          s[len-1] = '\0';
        }
    }
  free (c);
  return s;
}

static int
dump_header (struct xges_create_details_t *t, FILE *fp)
{
  time_t now = time (NULL);
  struct tm tm;
  localtime_r (&now, &tm);
  char *cmd = munge_commandline (t->commandline);
  char timestamp[1024];
  strftime (timestamp, sizeof (timestamp), "%F %T", &tm);
  fprintf (fp, 
           "<ges version='0.1'>\n"
           "<!-- Generated with %s %s on %s with command:\n%s %s\n"
           "-->\n"
           "  <project properties='properties;' metadatas='metadatas, name=(string)&quot;New\\ GES\\ \\nCreate\\ Project&quot;, author=(string)%s, render-scale=(double)%-10.7f, year=(string)%d;'>\n"
           "    <encoding-profiles>\n"
           "      <encoding-profile name='pitivi-profile' description='Pitivi encoding profile' type='container' preset-name='webmmux' format='video/webm' >\n"
           "        <stream-profile parent='pitivi-profile' id='0' type='video' presence='0' format='video/x-vp8, profile=(string){ 0, 1, 2, 3 }' preset-name='vp8enc' restriction='video/x-raw, width=(int)%d, height=(int)%d, framerate=(fraction)%d/%d, pixel-aspect-ratio=(fraction)1/1' pass='0' variableframerate='0' />\n"
           "        <stream-profile parent='pitivi-profile' id='1' type='audio' presence='0' format='audio/x-vorbis, rate=(int)[ 1, 200000 ], channels=(int)[ 1, 255 ]' preset-name='vorbisenc' restriction='audio/x-raw, channels=(int)2, rate=(int)44100' />\n"
           "      </encoding-profile>\n"
           "    </encoding-profiles>\n", 
           PACKAGE, VERSION, timestamp, PACKAGE, cmd, getenv ("USER"), t->zoom, tm.tm_year + 1900, t->width, t->height, 
           t->framerate_numerator, t->framerate_denominator);
  free (cmd);
  return 0;
}


static void
cache_media_info (struct clip_details_t *c)
{
  c->cached_uri = get_media_uri (c->source_filename);
  c->cached_is_image = is_image (c->source_filename);
  if (c->cached_is_image)
    {
      c->cached_media_type = 4;
      c->cached_duration = ULONG_MAX;
      /* we have to give a duration to images
       * because they don't have a built-in duration */
      if (c->duration == 0.0)
        c->duration = 4.0;
    }
  else
    {
      c->cached_duration = get_media_duration (c->cached_uri);
      c->cached_media_type = get_media_type (c->cached_uri);
    }
  if (c->cached_media_type & 0x04)
    get_media_dimensions (c->cached_uri, &c->cached_width, &c->cached_height);
}

static int
dump_asset (struct clip_details_t *c, FILE *fp, char **uris, size_t *uris_len)
{
  char *uri = NULL;
  if (c->cached_uri == NULL)
    return 1;

  if (!c->cached_is_image)
    {
      if (c->cached_duration == 0)
        return 2;
      if (c->cached_media_type == 0)
        return 3;
    }
  if (c->greenscreen)
    {
      char *effect = "frei0r-filter-bluescreen0r";
      int found = 0;
      uri = NULL;
      while (uri = argz_next (*uris, *uris_len, uri))
        if (strcmp (uri, effect) == 0)
          found = 1;
      if (!found)
        {
          argz_add (uris, uris_len, effect);
          fprintf (fp, "      <asset id='%s' extractable-type-name='GESEffect' properties='properties, track-type=(int)1;' metadatas='metadatas;' />\n", effect);
        }
    }
  if (c->transition_name)
    {
      int found = 0;
      uri = NULL;
      while (uri = argz_next (*uris, *uris_len, uri))
        if (strcmp (uri, c->transition_name) == 0)
          found = 1;
      if (!found)
        {
          argz_add (uris, uris_len, c->transition_name);
          fprintf (fp, "      <asset id='%s' extractable-type-name='GESTransitionClip' properties='properties;' metadatas='metadatas;' />\n", c->transition_name);
        }
    }

  uri = NULL;
  while (uri = argz_next (*uris, *uris_len, uri))
    if (strcmp (uri, c->cached_uri) == 0)
      return 0;
  argz_add (uris, uris_len, c->cached_uri);
  fprintf (fp, "      <asset id='%s' extractable-type-name='GESUriClip' properties='properties, supported-formats=(int)%d, duration=(guint64)%llu;' metadatas='metadatas;' />\n", c->cached_uri, c->cached_media_type, c->cached_duration);

  return 0;
}

static int
dump_resources (struct xges_create_details_t *t, FILE *fp)
{
  int i, retval;
  fprintf (fp, "    <ressources>\n");
  char *uris = NULL;
  size_t uris_len = 0;
  for (i = 0; i < t->num_clips; i++)
    {
      retval = dump_asset (t->clips[i], fp, &uris, &uris_len);
      if (retval != 0)
        break;
    }
  free (uris);
  if (retval == 0)
    fprintf (fp, "    </ressources>\n");
  return retval;
}

static void
fill_vars (struct clip_details_t *c, unsigned long long int *in, unsigned long long int *out, unsigned long long int *at, unsigned long long int *current_position)
{
  *in = c->start * GST_SECOND;
  if (c->duration == 0)
    *out = c->cached_duration;
  else
    {
      *out = (c->start + c->duration) * GST_SECOND;
      if (*out > c->cached_duration)
        *out = c->cached_duration;
    }

  if (c->relative_position)
    {
      if (c->relative_position < 0)
        {
          double pos = *current_position / (double)GST_SECOND;
          if (c->relative_position * -1 > pos)
            *at = 0;
          else
            *at = *current_position + (c->relative_position * GST_SECOND);
        }
      else
        *at = *current_position + (c->relative_position * GST_SECOND);
    }
  else
    {
      if (c->position >= 0)
        *at = c->position * GST_SECOND;
      else
        *at = *current_position;
    }
  if (c->transition_overlap)
    *at -= c->transition_overlap * GST_SECOND;

  *current_position = *at + *out - *in;
  c->cached_in = *in;
  c->cached_out = *out;
  c->cached_at = *at;
}

static unsigned long long int
calculate_total_duration (struct xges_create_details_t *t)
{
  int i, j;
  unsigned long long int current_position;
  unsigned long long int total_duration = 0;
  for (j = 0; j < t->num_layers; j++)
    {
      current_position = 0;
      for (i = 0; i < t->num_clips; i++)
        {
          struct clip_details_t *c = t->clips[i];
          if (c->layer_number == j)
            {
              unsigned long long int in, out, at = 0;
              fill_vars (c, &in, &out, &at, &current_position);
            }
        }
      if (current_position > total_duration)
        total_duration = current_position;
    }
  return total_duration;
}

static int
dump_interp (double fade_in, double fade_out, double max, unsigned long long int in, unsigned long long int out, FILE *fp)
{
  int i;
  double levels[4];
  unsigned long long int points[4];
  int num_points = 0;
  if (fade_in)
    {
      levels[num_points] = 0.0;
      points[num_points] = 0;
      num_points++;
    }
  levels[num_points] = max;
  points[num_points] = (fade_in * GST_SECOND) + in;
  num_points++;
  levels[num_points] = max;
  points[num_points] = out-(fade_out * GST_SECOND);
  num_points++;
  if (fade_out)
    {
      levels[num_points] = 0.0;
      points[num_points] = out;
      num_points++;
    }
  for (i = 0; i < num_points; i++)
    {
      fprintf (fp, "%llu:", points[i]);
      char spaces[3];
      snprintf (spaces, sizeof (spaces), "  ");
      if (num_points - 1 == i)
        spaces[0]='\0';
      if (levels[i] == 1.0)
        fprintf (fp, "1%s", spaces);
      else if (levels[i] == 0.0)
        fprintf (fp, "0%s", spaces);
      else
        fprintf (fp, "%lf%s", levels[i], spaces);
    }
  return 0;
}

static int
dump_clip (struct xges_create_details_t *t, struct clip_details_t *c, int *id, int layer_id, unsigned long long int *position, FILE *fp)
{
  int retval = 0;
  unsigned long long int in, out, at = 0;
  fill_vars (c, &in, &out, &at, position);
  if (c->transition_name)
    {
      unsigned long long int transition_duration = c->transition_overlap * GST_SECOND;
      fprintf (fp, "        <clip id='%d' asset-id='%s' type-name='GESTransitionClip' layer-priority='0' track-types='4' start='%llu' duration='%llu' inpoint='0' rate='0' properties='properties;' >\n", *id, c->transition_name, at, transition_duration);
      fprintf (fp, "        </clip>\n");
      *id += 1;
    }
  fprintf (fp, "        <clip id='%d' asset-id='%s' type-name='GESUriClip' layer-priority='%d' track-types='%d' start='%llu' duration='%llu' inpoint='%llu' rate='0' properties='properties, mute=(boolean)false, is-image=(boolean)%s;' >\n", *id, c->cached_uri, layer_id, c->cached_media_type, at, out - in, in, c->cached_is_image ? "true" : "false");
  if (c->greenscreen)
    {
      fprintf (fp, "          <effect asset-id='frei0r-filter-bluescreen0r' clip-id='%d' type-name='GESEffect' track-type='4' track-id='0' properties='properties, priority=(uint)2, active=(boolean)true, track-type=(int)4;' metadatas='metadatas;' children-properties='properties, distance=(double)%lf, color-r=(float)0, color-g=(float)%lf, qos=(boolean)true, color-b=(float)0;'> \n", *id, c->greenscreen_distance / 100.0, c->greenscreen_threshold / 100.0);
      fprintf (fp, "          </effect>\n");

    }
  if (c->crop)
    {
      unsigned int top = c->crop_y;
      unsigned int bottom = c->crop_y + c->crop_height;
      unsigned int left = c->crop_x;
      unsigned int right = c->crop_x + c->crop_width;
      fprintf (fp, "          <effect asset-id='videocrop' clip-id='%d' type-name='GESEffect' track-type='4' track-id='0' properties='properties, active=(boolean)true, track-type=(int)4;' metadatas='metadatas;' children-properties='properties, bottom=(int)%u, left=(int)%u, qos=(boolean)true, right=(int)%u, top=(int)%u;'> \n", *id, bottom, left, right, top);
      fprintf (fp, "          </effect>\n");

    }
  if (c->cached_media_type & 0x02 && c->ignore_audio == 0)
    {
      struct layer_details_t *l = (struct layer_details_t*) t->layers[c->layer_number];
      if (l->mute)
        fprintf (fp, "          <source track-id='1' children-properties='properties, mute=(boolean)true, volume=(double)0;'/>\n");
      else
        {
          fprintf (fp, "          <source track-id='1' children-properties='properties, mute=(boolean)false, volume=(double)1;'>\n");
          fprintf (fp, "            <binding type='direct' source_type='interpolation' property='volume' mode='1' track_id='1' values =' ");
          if (c->volume < 0 || c->no_volume_fade)
            dump_interp (c->fade_in, c->fade_out, 0.1, in, out, fp);
          else
            dump_interp (c->fade_in, c->fade_out, c->volume /100.0, in, out, fp);
          fprintf (fp, " '/>\n");
          fprintf (fp, "          </source>\n");
        }
    }
  if (c->cached_media_type & 0x04)
    {
      int width = c->cached_width;
      if (c->width != 0)
        width = c->width;
      int height = c->cached_height;
      if (c->height != 0)
        height = c->height;
      fprintf (fp, "          <source track-id='0' children-properties='properties, alpha=(double)1, height=(int)%d, posx=(int)%ld, posy=(int)%ld, width=(int)%d;'>\n", height, c->posx, c->posy, width);
      fprintf (fp, "            <binding type='direct' source_type='interpolation' property='alpha' mode='1' track_id='0' values =' ");
      
      dump_interp (c->fade_in, c->fade_out, 1, in, out, fp);
      fprintf (fp, " '/>\n");
      fprintf (fp, "          </source>\n");
    }
  fprintf (fp, "        </clip>\n");
  return retval;
}

static int
dump_layer (struct xges_create_details_t *t, struct layer_details_t *l, int layer_id, FILE *fp)
{
  unsigned long long int current_position = 0;
  int retval, count = 0, i;
  fprintf (fp, "      <layer priority='%d' properties='properties, auto-transition=(boolean)true;' metadatas='metadatas, volume=(float)%d;'>\n", l->id, l->mute ? 0 : 1);
  for (i = 0; i < t->num_clips; i++)
    {
      struct clip_details_t *c = t->clips[i];
      if (c->layer_number == l->id)
        {
          retval = dump_clip (t, t->clips[i], &count, layer_id, 
                              &current_position, fp);
          count++;
          if (retval != 0)
            break;
        }
    }
  if (retval == 0)
    fprintf (fp, "      </layer>\n");
  return retval;
}

static int
dump_timeline (struct xges_create_details_t *t, FILE *fp)
{
  int i, retval;
  unsigned long long int total_duration = calculate_total_duration (t);
  fprintf (fp, "    <timeline properties='properties, auto-transition=(boolean)true, snapping-distance=(guint64)166480473;' metadatas='metadatas, duration=(guint64)%llu;'>>\n", total_duration);
  fprintf (fp, "      <track caps='video/x-raw' track-type='4' track-id='0' properties='properties, async-handling=(boolean)false, message-forward=(boolean)false, caps=(string)video/x-raw, restriction-caps=(string)&quot;video/x-raw\\,\\ width\\=\\(int\\)%d\\,\\ height\\=\\(int\\)%d\\,\\ framerate\\=\\(fraction\\)%d/%d&quot;;' metadatas='metadatas;'/>\n", t->width, t->height, t->framerate_numerator, t->framerate_denominator);
  fprintf (fp, "      <track caps='audio/x-raw' track-type='2' track-id='1' properties='properties, async-handling=(boolean)false, message-forward=(boolean)false, caps=(string)audio/x-raw, restriction-caps=(string)&quot;NULL&quot;;' metadatas='metadatas;'/>\n");
  for (i = 0; i < t->num_layers; i++)
    {
      retval = dump_layer (t, t->layers[i], i, fp);
      if (retval != 0)
        break;
    }
  if (retval == 0)
    {
      fprintf (fp, "      <groups>\n");
      fprintf (fp, "      </groups>\n");
      fprintf (fp, "    </timeline>\n");
    }
  return retval;
}

static int
get_largest_width (struct xges_create_details_t *t)
{
  int i, max_width = 0;
  for (i = 0; i < t->num_clips; i++)
    {
      int w = clip_get_width (t->clips[i]);
      if (w > max_width)
        max_width = w;
    }
  return max_width;
}

static int
get_largest_height (struct xges_create_details_t *t)
{
  int i, max_height = 0;
  for (i = 0; i < t->num_clips; i++)
    {
      int h = clip_get_height (t->clips[i]);
      if (h > max_height)
        max_height = h;
    }
  return max_height;
}

static void
do_promotions (struct xges_create_details_t *t)
{
  int i;
  for (i = 0; i < t->num_promotions; i++)
    clip_promote (t->promotions[i], &t->clips, &t->num_clips, t->num_layers);
}

static void
do_splices (struct xges_create_details_t *t)
{
  int i;
  for (i = 0; i < t->num_clips; i++)
    {
      if (clip_has_splice (t->clips[i]))
        {
          clip_splice (t->clips[i], &t->clips, &t->num_clips);
          i = 0;
        }
    }
}

static void
dump_clip_info (clip_t c, FILE *fp)
{
  //in json, why not
  fprintf (fp, " \"clip\": {\n  ");
  char *file = strrchr (clip_get_uri (c), '/');
  if (file)
    file++;
  fprintf (fp, "\"uri\": \"%s\", \"layer\": \"%d\", \"start\": \"%lf\", \"duration\": \"%lf\"", 
           file, clip_get_layer_number (c), clip_get_start (c), 
           clip_get_duration (c) != 0 ? clip_get_duration (c) : (double)clip_get_cached_duration (c) / GST_SECOND);
  if (clip_has_position (c))
    fprintf (fp, ", \"position\": \"%lf\"", clip_get_position (c));
  if (clip_has_relative_position (c))
    fprintf (fp, ", \"relative-position\": \"%lf\"", clip_get_relative_position (c));
  if (clip_has_fade (c))
    fprintf (fp,", \"fade-in\": \"%lf\" \"fade-out\": \"%lf\"", clip_get_fade_in (c), clip_get_fade_out (c));
  if (clip_has_crop (c))
    fprintf (fp, ", \"crop\": \"%dx%d %d,%d\"", clip_get_crop_width (c), clip_get_crop_height (c), clip_get_crop_x (c), clip_get_crop_y (c));
  if (clip_has_zoom (c))
    fprintf (fp, ", \"zoom\": \"%d,%d %lf%%\"", clip_get_zoom_x (c), clip_get_zoom_y (c), clip_get_zoom_percent (c));
  if (clip_get_greenscreen (c))
    fprintf (fp, ", \"greenscreen\": \"1\"");
  if (clip_has_transition (c))
    fprintf (fp, ", \"transition\": \"%s %lf\"", clip_get_transition_name (c), clip_get_transition_duration (c));
  if (clip_is_audio_only (c))
    fprintf (fp, ", \"audio-only\": \"1\"");
  fprintf (fp, "\n");
  fprintf (fp, " },\n");
}

static void
dump_clips_info (clip_t *clips, int num_clips, FILE *fp)
{
  fprintf (fp, "clips: {\n");
  int i;
  for (i = 0; i < num_clips; i++)
    dump_clip_info (clips[i], fp);
  fprintf (fp, "}\n");
}

static int
generate_xges_file (struct xges_create_details_t *t, FILE *fp)
{
  int i;
  for (i = 0; i < t->num_clips; i++)
    cache_media_info (t->clips[i]);
  do_splices (t);
  for (i = 0; i < t->num_promotions; i++)
    {
      if (promote_get_layer (t->promotions[i]) >= t->num_layers)
        error (1, 0, "invalid layer `%d' given to " PROMOTE_CMD, 
               promote_get_layer (t->promotions[i]));
      if (promote_get_destination_layer (t->promotions[i]) >= t->num_layers)
        error (1, 0, "invalid argument `%d' given to --destination-layer" , 
               promote_get_layer (t->promotions[i]));
    }
  do_promotions (t);
  if (t->display)
    {
      display_project (t);
      if (!t->dump_clips)
        exit (0);
    }
  for (i = 0; i < t->num_clips; i++)
    {
      if (clip_has_crop (t->clips[i]))
        if (!clip_crop_extents_in_bounds (t->clips[i]))
            error (1, 0, "invalid bounds given for crop");
      if (clip_has_fade (t->clips[i]))
        if (!clip_fade_in_bounds (t->clips[i]))
          error (1, 0, "invalid bounds given for fade");
    }
  for (i = 0; i < t->num_clips; i++)
    {
      if (clip_has_zoom (t->clips[i]))
        {
          if (!clip_zoom_point_in_bounds (t->clips[i]))
            error (1, 0, "invalid bounds given for zoom");
          clip_fill_zoom (t->clips[i]);
        }
    }
  if (t->dump_clips)
    {
      dump_clips_info (t->clips, t->num_clips, stdout);
      exit (0);
    }
  if (t->width == 0)
    t->width = get_largest_width (t);
  if (t->height == 0)
    t->height = get_largest_height (t);
  int retval = dump_header (t, fp);
  if (retval == 0)
    retval = dump_resources (t, fp);
  if (retval == 0)
    dump_timeline (t, fp);
  if (retval == 0)
    dump_footer (fp);
  return retval;
}

int
generate_xges (struct xges_create_details_t *t)
{
  FILE *fp = stdout;
  if (t->output_filename)
    fp = fopen (t->output_filename, "w");
  int retval = generate_xges_file (t, fp);
  if (t->output_filename)
    fclose (fp);
  return retval;
}

static void
display_layer (struct xges_create_details_t *t, struct layer_details_t *l, unsigned long long int total_duration, int media_type)
{
  int i;
  char line[t->display_width+1], middle[t->display_width + 1];
  memset (line, ' ', sizeof (line));
  line[t->display_width - 1] = '\n';
  line[t->display_width] = '\0';
  memcpy (middle, line, sizeof (line));
  char *name = NULL;
  if (media_type == 4)
    name = "video";
  else if (media_type == 2)
    name = "audio";
  else
    name = "     ";
  int count = 0;
  for (i = 0; i < t->num_clips; i++)
    {
      struct clip_details_t *c = t->clips[i];
      if (c->layer_number == l->id && c->cached_media_type & media_type)
        {
          count++;
          int start = (double)c->cached_at / (double)total_duration * (t->display_width - 2);
          int end = (double)(c->cached_at + c->cached_out - c->cached_in)/ (double)total_duration * (t->display_width - 2);
          if (end > t->display_width-2) //benfix, why does this happen
            end = t->display_width-2;
          int j;
          for (j = start; j < end; j++)
            if (line[j] != '+')
              line[j] = '-';
          line[start] = '+';
          line[end] = '+';
          middle[start] = '|';
          middle[end] = '|';
        }
    }
  if (count > 0)
    {
      printf ("     %s", line);
      printf ("%s%s", name, middle);
      printf ("     %s", line);
    }
}

void
display_project (struct xges_create_details_t *t)
{
  int i;
  unsigned long long int total_duration = calculate_total_duration (t);
  for (i = 0; i < t->num_layers; i++)
    {
      display_layer (t, t->layers[i], total_duration, 4);
      display_layer (t, t->layers[i], total_duration, 2);
    }
}
